module MyLib  where
import Data.Graph
import Data.Array
import Data.Map(Map)
import qualified Data.Map as M

newtype Var = Var {vertex :: Vertex}
  deriving Show
data Const = Const Vertex (Maybe Vertex)
  deriving (Eq,Show,Ord)
data Sat = Variable Var
         | CTrue
         | CFalse
         | Or Sat Sat
         | And Sat Sat
         | Neg Sat
         deriving Show

data Expression = Add [SubExpression]
  deriving Show
data SubExpression = Mul Const Var
                   | EConst Const
                   deriving Show

type Formula = (Expression, Var)

type Coloring = Map Vertex Color

type Color = Int

orStuff :: [Sat] -> Sat
orStuff = foldr (\x acc -> Or x acc) CTrue

andStuff :: [Sat] -> Sat
andStuff = foldr (\x acc -> And x acc) CFalse

graphToSat :: Graph -> Int -> Sat
graphToSat graph numOfColors= undefined

vertexToFormula :: Graph -> Vertex -> Formula
vertexToFormula graph vert = (vertexToExpression graph vert,Var vert)

vertexToExpression :: Graph -> Vertex -> Expression
vertexToExpression graph vert = Add $ (EConst $ Const vert Nothing) : fmap (vertexToSubExpression vert) (graph ! vert)

vertexToSubExpression :: Vertex -> Vertex -> SubExpression
vertexToSubExpression fromv tov = Mul (Const fromv (Just tov)) (Var tov)

evaluateExpr :: Int -> Map Const Color -> Coloring -> Expression -> Color
evaluateExpr numOfColors constM colo (Add xs) = (sum $ fmap (evaluateSubExpression numOfColors constM colo) xs) `mod` numOfColors

evaluateSubExpression :: Int -> Map Const Color -> Coloring -> SubExpression -> Color
evaluateSubExpression numOfColors constM colo (EConst c) = constM M.! c
evaluateSubExpression numOfColors constM colo (Mul c v)  = (constM M.! c) * (colo M.! vertex v) `mod` numOfColors

constants :: Expression -> [Const]
constants (Add xs) = helper xs
  where
    helper [] = []
    helper ((EConst x) : xs) = x : helper xs
    helper ((Mul x _)  : xs) = x : helper xs

formulaToAssigments :: Int -> Formula -> Coloring -> [Map Const Color]
formulaToAssigments numOfColors formu col = filter isSatisfied assignments
  where
    assignments :: [Map Const Color]
    assignments = fmap (M.fromList . zip (constants . fst $ formu) . numToAssign numConstants numOfColors) [1..numOfColors^numConstants]
    numToAssign 0 _      _ = []
    numToAssign n colors x = x `mod` colors : numToAssign (n-1) colors (x `quot` colors)
    numConstants = length . constants . fst $ formu
    isSatisfied assign = evaluateExpr numOfColors assign col (fst formu) == col M.! vertex (snd formu)

vGuessesCorrect :: Int -> Graph -> Vertex -> Coloring -> Sat
vGuessesCorrect numOfColors graph vert coloring = undefined

numOfVars :: Int -> Int
numOfVars = ceiling . logBase 2 . fromIntegral

testGraph :: Graph
testGraph = array (0,2) [(0,[1]),(1,[0,2]),(2,[1])]

testColoring :: Coloring
testColoring = M.fromList [(0,2),(1,1),(2,0)]
